# Licensed under a 3-clause BSD style license - see LICENSE.rst
# Also see SPARK_LICENSE.rst
#
# This module contains SPARK parser by John Aycock.
#
# *** NO DOCSTRING ALLOWED IN THIS MODULE ***
# The parser builds its internal tables by reading the docstring,
# so you cannot put documentation here.
#

# STDLIB
import re

# ASTROPY
from astropy import log

# LOCAL
from .exceptions import ParserError, GenericASTTraversalPruningException

__all__ = ['GenericScanner', 'GenericParser', 'GenericASTBuilder',
           'GenericASTTraversal', 'GenericASTMatcher']

__version__ = 'SPARK-0.6.1'


def _namelist(instance):
    namelist, namedict, classlist = [], {}, [instance.__class__]
    for c in classlist:
        for b in c.__bases__:
            classlist.append(b)
        for name in dir(c):
            if name not in namedict:
                namelist.append(name)
                namedict[name] = 1
    return namelist


def _dump(tokens, states):  # pragma: no cover
    out_str = '\n'
    for i in range(len(states)):
        out_str += f'state {i}\n'
        for (lhs, rhs), pos, parent in states[i]:
            out_str += f"\t{lhs}::={' '.join(rhs[:pos])}.{' '.join(rhs[pos:])},{parent}"
        if i < len(tokens):
            out_str += f'\ntoken {tokens[i]}\n'
    log.info(out_str)


class GenericScanner:
    def __init__(self):
        pattern = self.reflect()
        self.re = re.compile(pattern, re.VERBOSE)

        self.index2func = {}
        for name, number in self.re.groupindex.items():
            self.index2func[number-1] = getattr(self, 't_' + name)

    def makeRE(self, name):
        doc = getattr(self, name).__doc__
        rv = f'(?P<{name[2:]:s}>{doc:s})'
        return rv

    def reflect(self):
        rv = []
        for name in _namelist(self):
            if name[:2] == 't_' and name != 't_default':
                rv.append(self.makeRE(name))

        rv.append(self.makeRE('t_default'))
        return '|'.join(rv)

    def error(self, s, pos):
        raise ParserError(f'Lexical error at position {pos}')

    def tokenize(self, s):
        pos = 0
        n = len(s)
        while pos < n:
            m = self.re.match(s, pos)
            if m is None:
                self.error(s, pos)

            groups = m.groups()
            for i in range(len(groups)):
                if groups[i] and i in self.index2func:
                    self.index2func[i](groups[i])
            pos = m.end()

    def t_default(self, s):
        r'( . | \n )+'
        pass


class GenericParser:
    def __init__(self, start):
        self.rules = {}
        self.rule2func = {}
        self.rule2name = {}
        self.collectRules()
        self.startRule = self.augment(start)
        self.ruleschanged = 1

    _START = 'START'
    _EOF = 'EOF'

    #
    #  A hook for GenericASTBuilder and GenericASTMatcher.
    #
    def preprocess(self, rule, func):
        return rule, func

    def addRule(self, doc, func):
        rules = doc.split()

        index = []
        for i in range(len(rules)):
            if rules[i] == '::=':
                index.append(i-1)
        index.append(len(rules))

        for i in range(len(index)-1):
            lhs = rules[index[i]]
            rhs = rules[index[i]+2:index[i+1]]
            rule = (lhs, tuple(rhs))

            rule, fn = self.preprocess(rule, func)

            if lhs in self.rules:
                self.rules[lhs].append(rule)
            else:
                self.rules[lhs] = [rule]
            self.rule2func[rule] = fn
            self.rule2name[rule] = func.__name__[2:]
        self.ruleschanged = 1

    def collectRules(self):
        for name in _namelist(self):
            if name[:2] == 'p_':
                func = getattr(self, name)
                doc = func.__doc__
                self.addRule(doc, func)

    def augment(self, start):
        #
        #  Tempting though it is, this isn't made into a call
        #  to self.addRule() because the start rule shouldn't
        #  be subject to preprocessing.
        #
        startRule = (self._START, (start, self._EOF))
        self.rule2func[startRule] = lambda args: args[0]
        self.rules[self._START] = [startRule]
        self.rule2name[startRule] = ''
        return startRule

    def makeFIRST(self):
        union = {}
        self.first = {}

        for rulelist in self.rules.values():
            for lhs, rhs in rulelist:
                if lhs not in self.first:
                    self.first[lhs] = {}

                if len(rhs) == 0:
                    self.first[lhs][None] = 1
                    continue

                sym = rhs[0]
                if sym not in self.rules:
                    self.first[lhs][sym] = 1
                else:
                    union[(sym, lhs)] = 1
        changes = 1
        while changes:
            changes = 0
            for src, dest in union.keys():
                destlen = len(self.first[dest])
                self.first[dest].update(self.first[src])
                if len(self.first[dest]) != destlen:
                    changes = 1

    #
    #  An Earley parser, as per J. Earley, "An Efficient Context-Free
    #  Parsing Algorithm", CACM 13(2), pp. 94-102.  Also J. C. Earley,
    #  "An Efficient Context-Free Parsing Algorithm", Ph.D. thesis,
    #  Carnegie-Mellon University, August 1968, p. 27.
    #

    def typestring(self, token):
        return None

    def error(self, token):
        raise ParserError(f'Syntax error at or near "{token}" token')

    def parse(self, tokens):
        tree = {}
        tokens.append(self._EOF)
        states = {0: [(self.startRule, 0, 0)]}

        if self.ruleschanged:
            self.makeFIRST()

        for i in range(len(tokens)):
            states[i+1] = []

            if states[i] == []:
                break
            self.buildState(tokens[i], states, i, tree)

        # _dump(tokens, states)

        if i < len(tokens)-1 or states[i+1] != [(self.startRule, 2, 0)]:
            del tokens[-1]
            self.error(tokens[i-1])
        rv = self.buildTree(tokens, tree, ((self.startRule, 2, 0), i+1))
        del tokens[-1]
        return rv

    def buildState(self, token, states, i, tree):
        needsCompletion = {}
        state = states[i]
        predicted = {}

        for item in state:
            rule, pos, parent = item
            lhs, rhs = rule

            #
            #  A -> a . (completer)
            #
            if pos == len(rhs):
                if len(rhs) == 0:
                    needsCompletion[lhs] = (item, i)

                for pitem in states[parent]:
                    if pitem is item:
                        break

                    prule, ppos, pparent = pitem
                    plhs, prhs = prule

                    if prhs[ppos:ppos+1] == (lhs,):
                        new = (prule,
                               ppos+1,
                               pparent)
                        if new not in state:
                            state.append(new)
                            tree[(new, i)] = [(item, i)]
                        else:
                            tree[(new, i)].append((item, i))
                continue

            nextSym = rhs[pos]

            #
            #  A -> a . B (predictor)
            #
            if nextSym in self.rules:
                #
                #  Work on completer step some more; for rules
                #  with empty RHS, the "parent state" is the
                #  current state we're adding Earley items to,
                #  so the Earley items the completer step needs
                #  may not all be present when it runs.
                #
                if nextSym in needsCompletion:
                    new = (rule, pos+1, parent)
                    olditem_i = needsCompletion[nextSym]
                    if new not in state:
                        state.append(new)
                        tree[(new, i)] = [olditem_i]
                    else:
                        tree[(new, i)].append(olditem_i)

                #
                #  Has this been predicted already?
                #
                if nextSym in predicted:
                    continue
                predicted[nextSym] = 1

                ttype = (token is not self._EOF and self.typestring(token) or
                         None)
                if ttype is not None:
                    #
                    #  Even smarter predictor, when the
                    #  token's type is known.  The code is
                    #  grungy, but runs pretty fast.  Three
                    #  cases are looked for: rules with
                    #  empty RHS; first symbol on RHS is a
                    #  terminal; first symbol on RHS is a
                    #  nonterminal (and isn't nullable).
                    #
                    for prule in self.rules[nextSym]:
                        new = (prule, 0, i)
                        prhs = prule[1]
                        if len(prhs) == 0:
                            state.append(new)
                            continue
                        prhs0 = prhs[0]
                        if prhs0 not in self.rules:
                            if prhs0 != ttype:
                                continue
                            else:
                                state.append(new)
                                continue
                        first = self.first[prhs0]
                        if None not in first and ttype not in first:
                            continue
                        state.append(new)
                    continue

                for prule in self.rules[nextSym]:
                    #
                    #  Smarter predictor, as per Grune &
                    #  Jacobs' _Parsing Techniques_.  Not
                    #  as good as FIRST sets though.
                    #
                    prhs = prule[1]
                    if (len(prhs) > 0 and prhs[0] not in self.rules and
                            token != prhs[0]):
                        continue
                    state.append((prule, 0, i))

            #
            #  A -> a . c (scanner)
            #
            elif token == nextSym:
                # assert new not in states[i+1]
                states[i+1].append((rule, pos+1, parent))

    def buildTree(self, tokens, tree, root):
        stack = []
        self.buildTree_r(stack, tokens, -1, tree, root)
        return stack[0]

    def buildTree_r(self, stack, tokens, tokpos, tree, root):
        (rule, pos, parent), state = root

        while pos > 0:
            want = ((rule, pos, parent), state)
            if want not in tree:
                #
                #  Since pos > 0, it didn't come from closure,
                #  and if it isn't in tree[], then there must
                #  be a terminal symbol to the left of the dot.
                #  (It must be from a "scanner" step.)
                #
                pos = pos - 1
                state = state - 1
                stack.insert(0, tokens[tokpos])
                tokpos = tokpos - 1
            else:
                #
                #  There's a NT to the left of the dot.
                #  Follow the tree pointer recursively (>1
                #  tree pointers from it indicates ambiguity).
                #  Since the item must have come about from a
                #  "completer" step, the state where the item
                #  came from must be the parent state of the
                #  item the tree pointer points to.
                #
                children = tree[want]
                if len(children) > 1:
                    child = self.ambiguity(children)
                else:
                    child = children[0]

                tokpos = self.buildTree_r(stack, tokens, tokpos, tree, child)
                pos = pos - 1
                (crule, cpos, cparent), cstate = child
                state = cparent

        lhs, rhs = rule
        result = self.rule2func[rule](stack[:len(rhs)])
        stack[:len(rhs)] = [result]
        return tokpos

    def ambiguity(self, children):
        #
        #  XXX - problem here and in collectRules() if the same
        #    rule appears in >1 method.  But in that case the
        #    user probably gets what they deserve :-)  Also
        #    undefined results if rules causing the ambiguity
        #    appear in the same method.
        #
        sortlist = []
        name2index = {}
        for i in range(len(children)):
            ((rule, pos, parent), index) = children[i]
            lhs, rhs = rule
            name = self.rule2name[rule]
            sortlist.append((len(rhs), name))
            name2index[name] = i
        sortlist.sort()
        outlist = list(map(lambda ab: ab[1], sortlist))
        return children[name2index[self.resolve(outlist)]]

    def resolve(self, list):
        #
        #  Resolve ambiguity in favor of the shortest RHS.
        #  Since we walk the tree from the top down, this
        #  should effectively resolve in favor of a "shift".
        #
        return list[0]


#  GenericASTBuilder automagically constructs a concrete/abstract syntax tree
#  for a given input.  The extra argument is a class (not an instance!)
#  which supports the "__setslice__" and "__len__" methods.
#
#  XXX - silently overrides any user code in methods.
#
class GenericASTBuilder(GenericParser):
    def __init__(self, AST, start):
        super(GenericASTBuilder, self).__init__(start)
        self.AST = AST

    def rebind(self, lhs):
        return lambda args, lhs=lhs, self=self: self.buildASTNode(args, lhs)

    def preprocess(self, rule, func):
        lhs, rhs = rule
        return rule, self.rebind(lhs)

    def buildASTNode(self, args, lhs):
        children = []
        for arg in args:
            if isinstance(arg, self.AST):
                children.append(arg)
            else:
                children.append(self.terminal(arg))
        return self.nonterminal(lhs, children)

    def terminal(self, token):
        return token

    def nonterminal(self, type, args):
        rv = self.AST(type)
        rv[slice(0, len(args))] = args
        return rv


#  GenericASTTraversal is a Visitor pattern according to Design Patterns.  For
#  each node it attempts to invoke the method n_<node type>, falling
#  back onto the default() method if the n_* can't be found.  The preorder
#  traversal also looks for an exit hook named n_<node type>_exit (no default
#  routine is called if it's not found).  To prematurely halt traversal
#  of a subtree, call the prune() method -- this only makes sense for a
#  preorder traversal.  Node type is determined via the typestring() method.
#
class GenericASTTraversal:
    def __init__(self, ast):
        self.ast = ast

    def typestring(self, node):
        return node.type

    def prune(self):
        raise GenericASTTraversalPruningException

    def preorder(self, node=None):
        if node is None:
            node = self.ast

        try:
            name = 'n_' + self.typestring(node)
            if hasattr(self, name):
                func = getattr(self, name)
                func(node)
            else:
                self.default(node)
        except GenericASTTraversalPruningException:
            return

        for kid in node:
            self.preorder(kid)

        name = name + '_exit'
        if hasattr(self, name):
            func = getattr(self, name)
            func(node)

    def postorder(self, node=None):
        if node is None:
            node = self.ast

        for kid in node:
            self.postorder(kid)

        name = 'n_' + self.typestring(node)
        if hasattr(self, name):
            func = getattr(self, name)
            func(node)
        else:
            self.default(node)

    def default(self, node):
        pass


#  GenericASTMatcher.  AST nodes must have "__getitem__" and "__cmp__"
#  implemented.
#
#  XXX - makes assumptions about how GenericParser walks the parse tree.
#
class GenericASTMatcher(GenericParser):
    def __init__(self, start, ast):
        super(GenericASTMatcher, self).__init__(start)
        self.ast = ast

    def rebind(self, func):
        return lambda args, func=func, self=self: self.foundMatch(args, func)

    def preprocess(self, rule, func):
        lhs, rhs = rule
        rhslist = list(rhs)
        rhslist.reverse()

        return (lhs, tuple(rhslist)), self.rebind(func)

    def foundMatch(self, args, func):
        func(args[-1])
        return args[-1]

    def match_r(self, node):
        self.input.insert(0, node)
        children = 0

        for child in node:
            if children == 0:
                self.input.insert(0, '(')
            children = children + 1
            self.match_r(child)

        if children > 0:
            self.input.insert(0, ')')

    def match(self, ast=None):
        if ast is None:
            ast = self.ast
        self.input = []

        self.match_r(ast)
        self.parse(self.input)

    def resolve(self, list):
        #
        #  Resolve ambiguity in favor of the longest RHS.
        #
        return list[-1]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""``stsynphot`` configurable items.

The default configuration heavily depends on STScI TRDS structure
but it can be easily re-configured as the user wishes via
`astropy.config`.

``PYSYN_CDBS`` must be a defined system environment variable for
directories to be configured properly. It also overwrites
``synphot`` configurable items.

"""
# STDLIB
import os

# THIRD-PARTY
import numpy as np
from astropy import log
from astropy.config import ConfigNamespace, ConfigItem

# SYNPHOT
from synphot.config import Conf as synconf
from synphot.utils import generate_wavelengths

__all__ = ['conf', 'getref', 'showref', 'overwrite_synphot_config']


class Conf(ConfigNamespace):
    """Configuration parameters."""

    # Set up default wavelength
    _wave, _wave_str = generate_wavelengths(
        minwave=500, maxwave=26000, num=10000, delta=None, log=True,
        wave_unit='angstrom')

    # Root directory
    rootdir = ConfigItem(
        os.environ.get('PYSYN_CDBS', '/grp/redcat/trds'),
        'TRDS data root directory')

    # Graph, optical component, and thermal component tables
    graphtable = ConfigItem('mtab$*_tmg.fits', 'Graph table')
    comptable = ConfigItem('mtab$*_tmc.fits', 'Component table')
    thermtable = ConfigItem('mtab$*_tmt.fits', 'Thermal table')

    # Default wavelength in Angstrom and its description
    waveset_array = ConfigItem(
        _wave.value.tolist(),
        'Default wavelength set in Angstrom', 'float_list')
    waveset = ConfigItem(_wave_str, 'Default wavelength set description')

    # Telescope primary mirror collecting area in cm^2
    area = ConfigItem(45238.93416, 'Telescope collecting area in cm^2')

    # Common filter name
    clear_filter = ConfigItem('clear', 'Name for a clear filter')

    # Wavelength catalog file
    wavecatfile = ConfigItem(
        'synphot$wavecats/wavecat.dat', 'Wavelength catalog file')

    # Detector parameters file
    detectorfile = ConfigItem(
        'synphot$detectors.dat', 'Detector parameters file')

    # IRAF shortcuts file for stsynphot.stio.irafconvert()
    irafshortcutfile = ConfigItem(
        'synphot$irafshortcuts.txt',
        'col1=shortcut_name col2=relpath_to_rootdir, has header.')

    # Clean up
    del _wave
    del _wave_str


def _get_synphot_cfgitems():
    """Iterator for ``synphot`` configuration items."""
    for c in synconf.__dict__.values():
        if isinstance(c, ConfigItem):
            yield c


def overwrite_synphot_config(root):
    """Silently overwrite ``synphot`` configurable items to point to
    given root directory.

    Parameters
    ----------
    root : str
        Root directory name.

    """
    subdir_keys = ['calspec', 'extinction', 'nonhst']

    # Need this for Windows support
    if root.startswith(('http', 'ftp')):
        sep = '/'
    else:
        sep = os.sep  # Can be / or \

    for cfgitem in _get_synphot_cfgitems():
        path, fname = os.path.split(cfgitem())

        i = np.where(list(map(path.__contains__, subdir_keys)))[0]
        if len(i) == 0:
            continue

        subdir = subdir_keys[i[0]]

        if subdir == 'nonhst':
            cfgval = sep.join([root, 'comp', subdir, fname])
        else:
            cfgval = sep.join([root, subdir, fname])

        cfgitem.set(cfgval)


conf = Conf()

# Override SYNPHOT configuration
overwrite_synphot_config(conf.rootdir)


def _get_ref_cfgitems():
    """Iterator for configuration items to be displayed."""
    from .stio import get_latest_file, irafconvert

    for cfgitem, do_conv in (
            (Conf.graphtable, True),
            (Conf.comptable, True),
            (Conf.thermtable, True),
            (Conf.area, False),
            (Conf.waveset, False)):
        val = cfgitem()
        if do_conv:
            val = get_latest_file(irafconvert(val))
        yield cfgitem.name, val


def getref():
    """Return current values of select configurable items as a dictionary.

    Returns
    -------
    refdict : dict

    """
    return dict([x for x in _get_ref_cfgitems()])


def showref():  # pragma: no cover
    """Show the values of select configurable items."""
    info_str = '\n'
    for x in _get_ref_cfgitems():
        info_str += f'{x[0]:10s}: {x[1]}\n'
    log.info(info_str)



"""This module contains STIS/FUVMAMA commissioning tests.
Adapted from ``astrolib/pysynphot/from_commissioning/stis/test*.py``.
"""

# STDLIB
import os
import shutil

# THIRD-PARTY
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from ..utils import CommCase

# Local test data
datafiles = ['earthshine.fits', 'el1215a.fits', 'el1302a.fits', 'el1356a.fits',
             'el2471a.fits', 'ngc1068_template.fits', 'Zodi.fits']


def setup_module(module):
    """Copy test data to working directory so ``parse_spec`` works
    properly for both ``stsynphot`` and ASTROLIB PYSYNPHOT."""
    for datafile in datafiles:
        src = get_pkg_data_filename(os.path.join('data', datafile))
        shutil.copyfile(src, datafile)


def teardown_module(module):
    """Clean up test data in working directory."""
    for datafile in datafiles:
        os.remove(datafile)


class Test1156(CommCase):
    obsmode = 'stis,fuvmama,25mama'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18.5,vegamag)'


class Test1157(Test1156):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),20,vegamag)'


class Test1158(Test1156):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1159(CommCase):
    obsmode = 'stis,fuvmama,e140h,c1416'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1160(CommCase):
    obsmode = 'stis,fuvmama,e140h,c1416,s02x02'
    spectrum = 'spec($PYSYN_CDBS/calspec/bd_28d4211_stis_001.fits)'


class Test1161(CommCase):
    obsmode = 'stis,fuvmama,e140m,c1425'
    spectrum = ('rn(spec(Zodi.fits),band(johnson,v),23.3,vegamag)+'
                '(spec(el1215a.fits)*0.2+spec(el1302a.fits)*0.01333333333+'
                'spec(el1356a.fits)*0.012+spec(el2471a.fits)*0.01)')


class Test1162(Test1161):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1163(Test1161):
    spectrum = ('spec(earthshine.fits)+rn(spec(Zodi.fits),band(johnson,v),'
                '23.3,vegamag)+(spec(el1215a.fits)+spec(el1302a.fits)+'
                'spec(el1356a.fits)+spec(el2471a.fits))*2.0')


class Test1164(CommCase):
    obsmode = 'stis,fuvmama,e140m,c1425,s02x006'
    spectrum = 'em(1425.0,0.043487548828125,1.0E-10,flam)'


class Test1165(Test1164):
    spectrum = 'em(1425.0,1.0,1.0E-10,flam)'


class Test1166(CommCase):
    obsmode = 'stis,fuvmama,e140m,c1425,s02x02'
    spectrum = 'rn(icat(k93models,11900,0.0,4.0),band(johnson,v),10,vegamag)'


class Test1167(Test1166):
    spectrum = 'rn(icat(k93models,11900,0.0,4.0),band(johnson,v),6,vegamag)'


class Test1168(Test1166):
    spectrum = 'rn(icat(k93models,11900,0.0,4.0),band(johnson,v),7,vegamag)'


class Test1169(Test1166):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),22,vegamag)'


class Test1170(Test1166):
    spectrum = 'spec($PYSYN_CDBS/calspec/bd_28d4211_stis_001.fits)'


class Test1171(CommCase):
    obsmode = 'stis,fuvmama,f25lya'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1172(Test1171):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1173(CommCase):
    obsmode = 'stis,fuvmama,f25nd3'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1174(Test1173):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1175(CommCase):
    obsmode = 'stis,fuvmama,f25ndq1'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1176(Test1175):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1177(CommCase):
    obsmode = 'stis,fuvmama,f25ndq3'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1178(Test1177):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1179(CommCase):
    obsmode = 'stis,fuvmama,f25qtz'
    spectrum = 'rn(icat(k93models,30000,0.0,4.0),band(johnson,v),26,vegamag)'


class Test1180(Test1179):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1181(CommCase):
    obsmode = 'stis,fuvmama,f25srf2'
    spectrum = 'rn(icat(k93models,30000,0.0,4.0),band(johnson,v),26,vegamag)'


class Test1182(Test1181):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1183(CommCase):
    obsmode = 'stis,fuvmama,g140l'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1184(Test1183):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1185(CommCase):
    obsmode = 'stis,fuvmama,g140l,s52x01'
    spectrum = 'rn(spec(ngc1068_template.fits),band(johnson,v),9,vegamag)'


class Test1186(CommCase):
    obsmode = 'stis,fuvmama,g140l,s52x2'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),13,vegamag)'


class Test1187(Test1186):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),14,vegamag)'


class Test1188(Test1186):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),14.1,vegamag)'


class Test1189(Test1186):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),27.5,vegamag)'


class Test1190(Test1186):
    spectrum = ('rn(spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits),'
                'band(johnson,v),12.77,vegamag)')


class Test1191(Test1186):
    spectrum = ('rn(spec($PYSYN_CDBS/calspec/grw_70d5824_stis_001.fits),'
                'band(johnson,v),10.516,vegamag)')


class Test1192(Test1186):
    spectrum = 'spec($PYSYN_CDBS/calspec/grw_70d5824_stis_001.fits)'


class Test1193(CommCase):
    obsmode = 'stis,fuvmama,g140m,c1567'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1194(CommCase):
    obsmode = 'stis,fuvmama,g140m,c1567,s52x2'
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1195(CommCase):
    obsmode = 'stis,g140l,fuvmama,s52x2'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test1196(Test1195):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test1197(Test1195):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test1198(Test1195):
    spectrum = 'el1215a.fits'
    force = 'extrap'


class Test1199(Test1195):
    spectrum = 'el1302a.fits'
    force = 'extrap'


class Test1200(Test1195):
    spectrum = 'el1356a.fits'
    force = 'extrap'


# Both ASTROLIB PYSYNPHOT and ``synphot`` report this that this combo
# gives DisjointError, so skipping this one.
# class Test1201(Test1195):
#    spectrum = 'el2471a.fits'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test tables.py module.

.. note::

    Uses local graph and component tables so that the tests
    are not affected by TRDS changes.

"""

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from .. import exceptions
from ..tables import GraphTable, CompTable


def test_custom_primarea():
    """Test reading non-default PRIMAREA from graph table."""
    gt = GraphTable(get_pkg_data_filename('data/tables_primarea_tmg.fits'))
    assert gt.primary_area.value == 100


class TestGraphTable:
    """Test graph table."""
    def setup_class(self):
        self.gt = GraphTable(get_pkg_data_filename('data/tables_tmg.fits'))

    def test_primarea(self):
        assert self.gt.primary_area is None

    @pytest.mark.parametrize(
        ('modes', 'innode', 'ans'),
        [(['acs'], 1, 20),
         (['wfc1'], 20, 30),
         (['acs'], 0, -1),
         (['foo'], 1, 100)])
    def test_next_node(self, modes, innode, ans):
        assert self.gt.get_next_node(modes, innode) == ans

    def test_comp(self):
        comp, thcomp = self.gt.get_comp_from_gt(
            ['wfc3', 'ir', 'f098m', 'mjd#'], 1)
        assert (set(comp) == set(
            ['wfc3_ir_mask', 'wfc3_ir_primary', 'clear', 'wfc3_ir_secondary',
             'wfc3_ir_fold', 'wfc3_pom', 'wfc3_ir_mir1', 'wfc3_ir_mir2',
             'wfc3_ir_csm', 'wfc3_ir_f098m_m', 'wfc3_ir_qe', 'wfc3_ir_cor',
             'wfc3_ir_rcp', 'wfc3_ir_win']))
        assert (set(thcomp) == set(
            ['wfc3_ir_mask', 'wfc3_ir_pads', 'wfc3_ir_primary', 'clear',
             'wfc3_ir_secondary', 'wfc3_ir_fold', 'wfc3_pom', 'wfc3_ir_mir1',
             'wfc3_ir_mir2', 'wfc3_ir_csm', 'wfc3_ir_win', 'wfc3_ir_qe',
             'wfc3_ir_rcp', 'wfc3_ir_f098m', 'wfc3_ir_wmring']))

    def test_comp_exceptions(self):
        # Invalid innode
        with pytest.raises(exceptions.UnusedKeyword):
            self.gt.get_comp_from_gt(['acs'], 0)

        # Invalid outnode
        with pytest.raises(exceptions.IncompleteObsmode):
            self.gt.get_comp_from_gt(['acs'], 1)


class TestCompTable:
    """Test optical and thermal component tables."""
    def setup_class(self):
        self.ct = CompTable(get_pkg_data_filename('data/tables_tmc.fits'))
        self.tt = CompTable(get_pkg_data_filename('data/tables_tmt.fits'))

    def test_optical_files(self):
        files = self.ct.get_filenames(['clear', None, '', 'wfc3_ir_mask'])
        np.testing.assert_array_equal(files[:3], 'clear')
        assert files[3].endswith('wfc3_ir_mask_001_syn.fits')

    def test_thermal_files(self):
        files = self.tt.get_filenames(['wfc3_ir_wmring'])
        assert files[0].endswith('wfc3_ir_wmring_001_th.fits')

    def test_exceptions(self):
        with pytest.raises(exceptions.GraphtabError):
            self.ct.get_filenames(['foo'])



"""This module contains STIS/NUVMAMA commissioning tests.
Adapted from ``astrolib/pysynphot/from_commissioning/stis/test*.py``.
"""

# STDLIB
import os
import shutil

# THIRD-PARTY
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from ..utils import CommCase

# Local test data
datafiles = ['earthshine.fits', 'el1215a.fits', 'el1302a.fits', 'el1356a.fits',
             'el2471a.fits', 'HS20270651.dat', 'Zodi.fits']


def setup_module(module):
    """Copy test data to working directory so ``parse_spec`` works
    properly for both ``stsynphot`` and ASTROLIB PYSYNPHOT."""
    for datafile in datafiles:
        src = get_pkg_data_filename(os.path.join('data', datafile))
        shutil.copyfile(src, datafile)


def teardown_module(module):
    """Clean up test data in working directory."""
    for datafile in datafiles:
        os.remove(datafile)


class Test1202(CommCase):
    obsmode = 'stis,g230l,nuvmama,s52x2'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test1203(Test1202):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test1204(Test1202):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test1214(CommCase):
    obsmode = 'stis,nuvmama,25mama'
    spectrum = 'rn(icat(k93models,30000,0.0,4.0),band(johnson,v),26,vegamag)'


class Test1215(Test1214):
    spectrum = 'rn(icat(k93models,5860,0.0,4.4),band(johnson,v),5,vegamag)'


class Test1216(Test1214):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1217(CommCase):
    obsmode = 'stis,nuvmama,e230h,c2263'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1218(CommCase):
    obsmode = 'stis,nuvmama,e230h,c2263,s02x02'
    spectrum = 'rn(bb(50000),band(johnson,v),10.516,vegamag)'


class Test1219(Test1218):
    spectrum = ('rn(icat(k93models,44500,0.0,5.0),band(johnson,v),10.516,'
                'vegamag)')


class Test1220(Test1218):
    spectrum = 'rn(pl(4000.0,-1.0,flam),band(johnson,v),10.516,vegamag)'


class Test1221(Test1218):
    spectrum = 'rn(pl(4000.0,0.0,flam),band(johnson,v),10.516,vegamag)'


class Test1222(Test1218):
    spectrum = ('rn(spec($PYSYN_CDBS/calspec/bd_28d4211_stis_001.fits),'
                'band(johnson,v),10.516,vegamag)')


class Test1223(Test1218):
    spectrum = ('rn(spec($PYSYN_CDBS/calspec/bd_28d4211_stis_001.fits),'
                'box(2000.0,1.0),1.0e-12,flam)')


class Test1224(Test1218):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),10.516,vegamag)'


class Test1225(Test1218):
    spectrum = 'spec($PYSYN_CDBS/calspec/bd_28d4211_stis_001.fits)'


class Test1226(CommCase):
    obsmode = 'stis,nuvmama,e230m,c1978'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1227(CommCase):
    obsmode = 'stis,nuvmama,e230m,c1978,s02x02'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18.5,vegamag)'


class Test1228(Test1227):
    spectrum = 'spec($PYSYN_CDBS/calspec/bd_28d4211_stis_001.fits)'


class Test1229(CommCase):
    obsmode = 'stis,nuvmama,f25ciii'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1230(Test1229):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1231(CommCase):
    obsmode = 'stis,nuvmama,f25cn182'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1232(Test1231):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1233(CommCase):
    obsmode = 'stis,nuvmama,f25cn270'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1234(Test1233):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1235(CommCase):
    obsmode = 'stis,nuvmama,f25mgii'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1236(Test1235):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1237(CommCase):
    obsmode = 'stis,nuvmama,f25nd5'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1238(Test1237):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1239(CommCase):
    obsmode = 'stis,nuvmama,f25ndq2'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1240(Test1239):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1241(CommCase):
    obsmode = 'stis,nuvmama,f25ndq4'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1242(Test1241):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1243(CommCase):
    obsmode = 'stis,nuvmama,f25qtz'
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),18,vegamag)'


class Test1244(Test1243):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),20,vegamag)'


class Test1245(Test1243):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1246(CommCase):
    obsmode = 'stis,nuvmama,f25srf2'
    spectrum = 'rn(icat(k93models,30000,0.0,4.0),band(johnson,v),26,vegamag)'


class Test1247(Test1246):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1248(CommCase):
    obsmode = 'stis,nuvmama,g230l'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1249(Test1248):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1250(CommCase):
    # Original test used gal1 but it is no longer supported, so we use gal3
    obsmode = 'stis,nuvmama,g230l,s52x2'
    spectrum = ('rn(icat(k93models,44500,0.0,5.0)*ebmvx(0.5,gal3),'
                'band(johnson,v),15,vegamag)')


class Test1251(Test1250):
    # Original test used lmc but it is no longer supported, so we use lmcavg
    spectrum = ('rn(icat(k93models,44500,0.0,5.0)*ebmvx(0.5,lmcavg),'
                'band(johnson,v),15,vegamag)')


class Test1252(Test1250):
    # Original test used smc but it is no longer supported, so we use smcbar
    spectrum = ('rn(icat(k93models,44500,0.0,5.0)*ebmvx(0.5,smcbar),'
                'band(johnson,v),15,vegamag)')


class Test1253(Test1250):
    # Original test used xgal but it is no longer supported, so we use xgalsb
    spectrum = ('rn(icat(k93models,44500,0.0,5.0)*ebmvx(0.5,xgalsb),'
                'band(johnson,v),15,vegamag)')


class Test1254(Test1250):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),24,vegamag)'


class Test1255(Test1250):
    spectrum = 'spec($PYSYN_CDBS/calspec/grw_70d5824_stis_001.fits)'


class Test1256(CommCase):
    obsmode = 'stis,nuvmama,g230m,c2818'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test1257(CommCase):
    obsmode = 'stis,nuvmama,g230m,c2818,s52x2'
    spectrum = 'spec($PYSYN_CDBS/calspec/agk_81d266_stis_001.fits)'


class Test1258(CommCase):
    obsmode = 'stis,nuvmama,prism'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test1259(CommCase):
    obsmode = 'stis,nuvmama,prism,s52x01'
    spectrum = 'spec(HS20270651.dat)'
    force = 'extrap'


class Test1260(Test1259):
    obsmode = 'stis,nuvmama,prism,s52x2'



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module handles ``stsynphot``-specific I/O for:

* FITS - See `astropy.io.fits`
* Basic ASCII - See `astropy.io.ascii`

"""
# STDLIB
import fnmatch
import re
import os
import sys
import warnings
from pathlib import Path

# THIRD-PARTY
import numpy as np

# ASTROPY
import astropy
from astropy import units as u
from astropy.io import ascii, fits
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.introspection import minversion

# SYNPHOT
from synphot import exceptions as synexceptions
from synphot import units

__all__ = ['resolve_filename', 'irafconvert', 'get_latest_file',
           'read_graphtable', 'read_comptable', 'read_catalog', 'read_wavecat',
           'read_waveset', 'read_detector_pars', 'read_interp_spec']

_irafconvpat = re.compile(r'\$(\w*)')
_irafconvdata = None

ASTROPY_LT_4_3 = not minversion(astropy, '4.3')

if ASTROPY_LT_4_3:
    from astropy.utils.data import _find_pkg_data_path as get_pkg_data_path
else:
    from astropy.utils.data import get_pkg_data_path


def resolve_filename(path, *args):
    """Resolve filename that could be URL or local file.

    Parameters
    ----------
    path : str
        Root directory or URL.

    args : tuple of str
        Any sub-path(s) and the actual filename.

    Returns
    -------
    reg_filename : str
        Resolved filename.

    """
    path_lc = path.lower()
    if path_lc.startswith(('ftp', 'http')):
        sep = '/'
        fname = sep.join(args)
        if path_lc.endswith(sep):
            reg_filename = path + fname
        else:
            reg_filename = path + sep + fname

        # Catch-all to ensure result is OS-independent.
        if sys.platform.startswith('win'):
            reg_filename = reg_filename.replace('\\', sep)
    else:
        reg_filename = str(Path(path) / os.sep.join(args))

    return reg_filename


def _iraf_decode(irafdir):
    """Decode IRAF dir shortcut."""
    from .config import conf  # Put here to avoid circular import error
    global _irafconvdata

    irafdir = irafdir.lower()

    if irafdir == 'synphot':  # Local data
        path = get_pkg_data_path('data')
    elif irafdir == 'crrefer':  # Root dir
        path = conf.rootdir
    else:  # Read from file
        # Avoid repeated I/O but do not load if not used
        if _irafconvdata is None:
            _irafconvdata = ascii.read(irafconvert(conf.irafshortcutfile))

        mask = _irafconvdata['IRAFNAME'] == irafdir
        if not np.any(mask):
            raise KeyError(f'IRAF shortcut {irafdir} not found in '
                           f'{conf.irafshortcutfile}.')
        relpath = os.path.normpath(_irafconvdata['RELPATH'][mask][0])
        path = os.path.join(conf.rootdir, relpath)

    return path


def irafconvert(iraf_filename, sep='$'):
    """Convert IRAF filename to regular filename.

    Acceptable IRAF formats:

    * ``$path/file`` - ``path`` assumed to be environment variable.
    * ``path$file`` - ``path`` is special IRAF shortcut for TRDS
      data directory (case-insensitive).

    Notes on special IRAF shortcut:

    * ``synphot`` points to software data directory.
    * ``crrefer`` points to ``stsynphot.config.conf.rootdir``.
    * Otherwise, decoded based on
      ``stsynphot.config.conf.irafshortcutfile``
      that must contain the following named columns:

        #. ``IRAFNAME`` - The shortcut for look-up. If multiple matches
           are found, the first match is used.
        #. ``RELPATH`` - Path relative to
           ``stsynphot.config.conf.rootdir``.

    If separator is not found, input is returned as-is.

    Parameters
    ----------
    iraf_filename : str
        IRAF filename.

    sep : char, optional
        Path-file separator.

    Returns
    -------
    reg_filename : str
        Regular filename.

    Raises
    ------
    KeyError
        Environment variable or IRAF shortcut is undefined.

    TypeError
        Input is not a string.

    """
    if not isinstance(iraf_filename, str):
        raise TypeError(f'{iraf_filename} is not a string.')

    # Nothing needs to be done
    if sep not in iraf_filename:
        return iraf_filename

    # Remove duplicate separators and extraneous relative paths.
    iraf_filename = os.path.normpath(iraf_filename)

    # $var/file
    if iraf_filename.startswith(sep):
        match = _irafconvpat.match(iraf_filename)
        path = os.environ[match.group(1)]
        fname = iraf_filename[match.end() + 1:]  # 1 to omit leading slash

    # dir$file
    else:
        irafdir, fname = iraf_filename.split(sep)
        path = _iraf_decode(irafdir)

    return resolve_filename(path, fname)


# TODO: Use CRDS instead.
def get_latest_file(template, raise_error=False, err_msg=''):
    """Find the filename that appears last in sorted order
    based on given template.

    Parameters
    ----------
    template : str
        Search template in the form of ``path/pattern``
        where pattern is acceptable by :py:mod:`fnmatch`.

    raise_error : bool, optional
        Raise an error when no files found.
        Otherwise, will issue warning only.

    err_msg : str
        Alternate message for when no files found.
        If not given, generic message is used.

    Returns
    -------
    filename : str
        Latest filename.

    Raises
    ------
    IOError
        No files found.

    """
    path, pattern = os.path.split(irafconvert(template))
    path_lc = path.lower()

    if not path:
        path = os.curdir

    # Remote HTTP/HTTPS directory
    if path_lc.startswith('http'):
        from urllib import request
        from bs4 import BeautifulSoup

        with request.urlopen(path) as fin:  # nosec
            soup = BeautifulSoup(fin, 'html.parser')

        allfiles = [x.text for x in soup.find_all('a')]
        sep = '/'

    # Remote FTP directory
    elif path_lc.startswith('ftp:'):
        from urllib import request

        response = request.urlopen(path).read().decode('utf-8').splitlines()  # nosec  # noqa
        allfiles = list(set([x.split()[-1] for x in response]))  # Rid symlink
        sep = '/'

    # Local directory
    elif os.path.isdir(path):
        allfiles = os.listdir(path)
        sep = os.sep

    # Bogus directory
    else:
        allfiles = []

    matched_files = sorted(fnmatch.filter(allfiles, pattern))

    # Last file in sorted listing
    if matched_files:
        filename = path + sep + matched_files[-1]

    # No files found
    else:
        if not err_msg:
            err_msg = f'No files found for {template}'

        if raise_error:
            raise IOError(err_msg)
        else:
            warnings.warn(err_msg, AstropyUserWarning)
            filename = ''

    return filename


def _read_table(filename, ext, dtypes):
    """Generic table reader.

    Parameters
    ----------
    filename : str
        Table filename.
        If suffix is not 'fits' or 'fit', assume ASCII format.

    ext : int
        Data extension.
        This is ignored for ASCII file.

    dtypes : dict
        Dictionary that maps column names to data types.

    Returns
    -------
    data : `~astropy.io.fits.FITS_rec` or `~astropy.table.Table`
        Data table.

    Raises
    ------
    synphot.exceptions.SynphotError
        Failure to parse table.

    """
    # FITS
    if filename.endswith(('.fits', '.fit')):
        with fits.open(filename) as f:
            data = f[ext].data.copy()

        err_str = ''
        for key, val in dtypes.items():
            if not np.issubdtype(data[key].dtype, val):
                err_str += (f'Expect {key} to be {val} but get '
                            f'{data[key].dtype}.\n')
        if err_str:
            raise synexceptions.SynphotError(err_str)

    # ASCII
    else:  # pragma: no cover
        converters = dict(
            [[k, ascii.convert_numpy(v)] for k, v in dtypes.items()])
        data = ascii.read(filename, converters=converters)

    return data


def read_graphtable(filename, tab_ext=1):
    """Read graph table file.

    Table must contain the following named columns:

    #. ``COMPNAME`` - Component name, usually filter name (str)
    #. ``KEYWORD`` - Usually instrument name (str)
    #. ``INNODE`` - Input node number (int)
    #. ``OUTNODE``- Output node number (int)
    #. ``THCOMPNAME`` - Thermal component name, usually filter name (str)
    #. ``COMMENT`` - Comment (str)

    Example:

    +--------+-------+------+-------+----------+--------+
    |COMPNAME|KEYWORD|INNODE|OUTNODE|THCOMPNAME|COMMENT |
    +========+=======+======+=======+==========+========+
    | clear  |nicmos |   1  |  30   |  clear   |idno=100|
    +--------+-------+------+-------+----------+--------+
    | clear  | wfc3  |   1  |  30   |  clear   |        |
    +--------+-------+------+-------+----------+--------+
    | clear  | wfpc  |   1  |  20   |  clear   |idno=100|
    +--------+-------+------+-------+----------+--------+

    Parameters
    ----------
    filename : str
        Graph table filename.
        If suffix is not 'fits' or 'fit', assume ASCII format.

    tab_ext : int, optional
        FITS extension index of the data table.
        This is ignored for ASCII file.

    Returns
    -------
    primary_area : `~astropy.units.quantity.Quantity` or `None`
        Value of PRIMAREA keyword in primary header.
        Always `None` for ASCII file.

    data : `~astropy.io.fits.FITS_rec` or `~astropy.table.Table`
        Data table.

    Raises
    ------
    synphot.exceptions.SynphotError
        Failure to parse graph table.

    """
    graph_dtypes = {
        'COMPNAME': np.str_, 'KEYWORD': np.str_, 'INNODE': np.int32,
        'OUTNODE': np.int32, 'THCOMPNAME': np.str_, 'COMMENT': np.str_}
    data = _read_table(filename, tab_ext, graph_dtypes)

    # Get primary area
    if filename.endswith(('.fits', '.fit')):
        with fits.open(filename) as f:
            primary_area = f[str('PRIMARY')].header.get('PRIMAREA', None)
    else:  # pragma: no cover
        primary_area = None

    if primary_area is not None and not isinstance(primary_area, u.Quantity):
        primary_area = primary_area * units.AREA

    # Check for segmented graph table
    if np.any([x.lower().endswith('graph')
               for x in data['COMPNAME']]):  # pragma: no cover
        raise synexceptions.SynphotError(
            'Segmented graph tables not supported.')

    return primary_area, data


def read_comptable(filename, tab_ext=1):
    """Read component table file (regular or thermal).

    Table must contain the following named columns:

    #. ``TIME`` - Useafter date and time in the format of
       ``MMM DD YYYY HH:MM:SS`` (str)
    #. ``COMPNAME`` - Component name, usually a combination of
       instrument/detector and filter names (str)
    #. ``FILENAME`` - Path to corresponding throughput file in the
       format of ``path_var$filename`` (str)
    #. ``COMMENT`` - Comment (str)

    See :ref:`stsynphot-master-comp` for more details.

    Parameters
    ----------
    filename : str
        Component table filename.
        If suffix is not 'fits' or 'fit', assume ASCII format.

    tab_ext : int, optional
        FITS extension index of the data table.
        This is ignored for ASCII file.

    Returns
    -------
    data : `~astropy.io.fits.FITS_rec` or `~astropy.table.Table`
        Data table.

    """
    return _read_table(filename, tab_ext,
                       {'TIME': np.str_, 'COMPNAME': np.str_,
                        'FILENAME': np.str_, 'COMMENT': np.str_})


def read_catalog(filename, tab_ext=1):
    """Read catalog grid look-up table.

    Table must contain the following named columns:

    #. ``INDEX`` - Grid values (str)
    #. ``FILENAME`` - Relative file path and column name (str)

    Example:

    +----------------+-----------------------------+
    | INDEX          | FILENAME                    |
    +================+=============================+
    | 10000,-0.5,0.0 | ckm05/ckm05_10000.fits[g00] |
    +----------------+-----------------------------+
    | 10000,-0.5,0.5 | ckm05/ckm05_10000.fits[g05] |
    +----------------+-----------------------------+
    | 10000,-0.5,1.0 | ckm05/ckm05_10000.fits[g10] |
    +----------------+-----------------------------+

    Parameters
    ----------
    filename : str
        Catalog filename.
        If suffix is not 'fits' or 'fit', assume ASCII format.

    tab_ext : int, optional
        FITS extension index of the data table.
        This is ignored for ASCII file.

    Returns
    -------
    data : `~astropy.io.fits.FITS_rec` or `~astropy.table.Table`
        Data table.

    """
    return _read_table(filename, tab_ext, {'INDEX': np.str_,
                                           'FILENAME': np.str_})


def read_wavecat(filename):
    """Read wavelength catalog from ASCII file.

    Table must contain two columns without header.
    Comment lines are allowed and will be ignored.
    Columns are automatically named:

    #. ``OBSMODE`` - Observation mode.
    #. ``FILENAME`` - Corresponding wavelength table filename
       or parameters.

    Example::

        # WAVECAT.DAT -- Comments.
        # More comments.
        cos,fuv     (900.0,3000.0,1.0)
        cos,nuv     (1000.0,12000.0,1.0)
        acs,hrc     synphot$wavecats/acs.dat
        acs,wfc1    synphot$wavecats/acs.dat

    Parameters
    ----------
    filename : str
        Wavelength catalog filename. Must be ASCII format.

    Returns
    -------
    data : `~astropy.table.Table`
        Data table.

    """
    return ascii.read(
        filename, names=('OBSMODE', 'FILENAME'),
        guess=False, format='no_header',
        converters={'OBSMODE': [ascii.convert_numpy(str)],
                    'FILENAME': [ascii.convert_numpy(str)]})


def read_waveset(filename, wave_unit=u.AA):
    """Read wavelength table from ASCII file.

    Table must contain a single column without header.
    Comment lines are allowed and will be ignored.
    Column is automatically named ``WAVELENGTH``.

    Example::

        # ACS.DAT -- Comments.
        # More comments.
        1000.
        2000.
        5000.
        9000.

    Parameters
    ----------
    filename : str
        Wavelength table filename. Must be ASCII format.

    wave_unit : str or `~astropy.units.Unit`
        Wavelength unit.

    Returns
    -------
    waveset : `~astropy.units.quantity.Quantity`
        Wavelength set array.

    """
    wave_unit = units.validate_wave_unit(wave_unit)
    data = ascii.read(
        filename, guess=False, format='no_header', names=('WAVELENGTH', ),
        converters={'WAVELENGTH': [ascii.convert_numpy(float)]})
    waveset = data['WAVELENGTH'].data

    if not isinstance(waveset, u.Quantity):
        waveset = waveset * u.AA

    return waveset


def read_detector_pars(filename):
    """Read detector parameters from ASCII file.

    Table must contain 4 columns without header.
    Comment lines are allowed and will be ignored.
    Columns are automatically named:

    #. ``OBSMODE`` - Observation mode.
    #. ``SCALE`` - Pixel scale in arcseconds.
    #. ``NX`` - X dimension in pixels.
    #. ``NY`` - Y dimension in pixels.

    Example::

        # DETECTORS.DAT -- Comments.
        # More comments.
        acs,hrc     0.027   1024  1024
        acs,sbc     0.032   1024  1024
        stis,g140l  0.0244  1024  1024
        stis,g140m  0.0290  1024  1024

    Parameters
    ----------
    filename : str
        Detector parameters filename. Must be ASCII format.

    Returns
    -------
    data : `~astropy.table.Table`
        Data table.

    """
    return ascii.read(
        filename, guess=False, format='no_header', delimiter=r'\s',
        names=('OBSMODE', 'SCALE', 'NX', 'NY'),
        converters={'OBSMODE': [ascii.convert_numpy(str)],
                    'SCALE': [ascii.convert_numpy(float)],
                    'NX': [ascii.convert_numpy(int)],
                    'NY': [ascii.convert_numpy(int)]})


def read_interp_spec(filename, tab_ext=1):
    """Read parameterized (interpolate-able) throughput
    spectra from FITS table.

    Table must contain two or more columns:

    #. ``WAVELENGTH`` - Wavelength values.
    #. ``PAR#VAL1`` - First parameterized column.
    #. ``PAR#VAL2`` - Second parameterized column.
    #. ...

    Example:

    +------------+---------------+---------------+
    | WAVELENGTH | FR656N#6274.0 | FR656N#6331.4 |
    +============+===============+===============+
    | 6136.0     | 0.0012254583  | 0.00019133692 |
    +------------+---------------+---------------+
    | 6141.0     | 0.0016151578  | 0.00021833261 |
    +------------+---------------+---------------+
    | 6146.0     | 0.0021577575  | 0.00025030275 |
    +------------+---------------+---------------+

    Parameters
    ----------
    filename : str
        FITS filename.

    tab_ext : int, optional
        FITS extension index of the data table.

    Returns
    -------
    data : `~astropy.io.fits.FITS_rec`
        Data table.

    wave_unit : str
        Value of ``TUNIT1`` in table header.

    do_wave_shift : bool
        Perform wavelength shift before interpolation.
        This is `True` when primary header has ``PARAMS``
        keyword set to ``WAVELENGTH`` (case-insensitive).

    allow_extrap : bool
        Allow extrapolation. This is only `True` when
        primary header has ``EXTRAP`` keyword explicitly
        set to ``T`` or `True`.

    """
    with fits.open(filename) as f:
        pri_hdr = f[str('PRIMARY')].header

        params = pri_hdr.get('PARAMS', '')
        if params.lower() == 'wavelength':
            do_wave_shift = True
        else:
            do_wave_shift = False

        extrap = pri_hdr.get('EXTRAP', False)
        if extrap in (True, 'T', 't'):
            allow_extrap = True
        else:
            allow_extrap = False

        wave_unit = f[tab_ext].header['TUNIT1'].lower()
        data = f[tab_ext].data.copy()

    return data, wave_unit, do_wave_shift, allow_extrap



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test spectrum.py module."""

# STDLIB
import os
import shutil
import tempfile
import warnings

# THIRD-PARTY
import numpy as np
import pytest

# ASTROPY
from astropy import units as u
from astropy.io import fits
from astropy.modeling.models import Const1D
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning

# SYNPHOT
from synphot import units
from synphot import exceptions as synexceptions
from synphot.models import ConstFlux1D
from synphot.spectrum import SourceSpectrum, SpectralElement

# LOCAL
from .. import spectrum
from ..config import conf
from ..exceptions import PixscaleNotFoundError
from ..stio import irafconvert

GT_FILE = get_pkg_data_filename(os.path.join('data', 'tables_tmg.fits'))
CP_FILE = get_pkg_data_filename(os.path.join('data', 'tables_tmc.fits'))
TH_FILE = get_pkg_data_filename(os.path.join('data', 'tables_tmt.fits'))


@pytest.mark.remote_data
class TestInterpolateSpectrum:
    """Test spectrum interpolation."""
    def setup_class(self):
        self.fname_acs = irafconvert(
            'cracscomp$acs_wfc_aper_002_syn.fits[aper#]')
        self.fname_cos = irafconvert(
            'crcoscomp$cos_mcp_g140lc1230_mjd_013_syn.fits[mjd#]')
        self.fname_ramp = irafconvert(
            'cracscomp$acs_fr656n_005_syn.fits[fr656n#]')
        self.fname_stis = irafconvert(
            'crstiscomp$stis_nm16_mjd_010_syn.fits[MJD#]')
        self.wave_stis = [1049, 2250, 3500, 4750, 6000, 7250, 8500, 9750,
                          11000]

    def test_no_interp_first_col(self):
        """Test whether the algorithm grabs the correct column
        when the input value is the first column (no interpolation required)
        and there is no THROUGHPUT (default) column present, meaning that the
        interpolation column starts at column 1.

        """
        sp = spectrum.interpolate_spectral_element(self.fname_acs, 0)
        w = [3500, 4000, 5000, 6000, 7000, 8000, 9000, 10000, 11000]
        np.testing.assert_allclose(sp.waveset.value, w)
        np.testing.assert_allclose(
            sp(w).value,
            [0.28, 0.22, 0.20999999, 0.22, 0.22, 0.2, 0.15000001, 0.1, 0.04])
        assert 'acs_wfc_aper_002_syn.fits#0' in sp.meta['expr']

    def test_no_interp_has_err_col(self):
        """Test whether the algorithm grabs the correct column
        when the input value is a column (no interpolation required) and
        an ERROR column is present.

        """
        sp = spectrum.interpolate_spectral_element(self.fname_stis, 51252)
        np.testing.assert_allclose(sp.waveset[::25].value, self.wave_stis)
        np.testing.assert_allclose(
            sp(sp.waveset[:10]).value,
            [0, 0.965065, 0.965065, 0.965065, 0.963328, 0.976245, 0.983497,
             0.984206, 0.976436, 0.963131])

    def test_interp_mjd(self):
        """Test whether the algorithm gets the correct throughput table
        when input value does not correspond to a column and interpolation
        is required.

        """
        sp = spectrum.interpolate_spectral_element(self.fname_stis, 51000)
        np.testing.assert_array_equal(sp.waveset[::25].value, self.wave_stis)
        np.testing.assert_allclose(
            sp(sp.waveset[:10]).value,
            [0, 0.97830353, 0.97830353, 0.97830353, 0.97722476, 0.98524689,
             0.98975077, 0.99019109, 0.98536552, 0.97710241])
        assert sp(sp.waveset[0]) == sp(sp.waveset[-1])

    @pytest.mark.parametrize(
        ('interpval', 'ans'),
        [(40000, [0, 0.10225279, 0.06328997, 0.02321319, 0.00536039, 0]),
         (60000, [0, 0.09322135, 0.04873522, 0.01732031, 0.00407352, 0])])
    def test_extrap_mjd(self, interpval, ans):
        """Test extrapolation without using default column."""
        sp = spectrum.interpolate_spectral_element(self.fname_cos, interpval)
        w = sp.waveset[400:3000:500]
        np.testing.assert_allclose(sp(w).value, ans, rtol=1e-5)

    def test_interp_ramp(self):
        """Test ramp filter interpolation with wavelength shift."""
        sp = spectrum.interpolate_spectral_element(self.fname_ramp, 6480)
        w = sp.waveset[:100:10]
        np.testing.assert_allclose(
            w.value,
            [3499.99902344, 5686, 5736, 5786, 5836, 5886, 5936, 5986, 6036,
             6086])
        np.testing.assert_allclose(
            sp(w).value,
            [5.88854514e-07, 1.00120149e-06, 1.00202179e-06, 1.00353753e-06,
             1.00710832e-06, 1.03335088e-06, 1.34637005e-06, 3.78240595e-06,
             1.53430929e-05, 5.09230156e-05])

    def test_default_ramp(self):
        """Test ramp filter using default THROUGHPUT with warning."""
        with pytest.warns(AstropyUserWarning,
                          match=r'Extrapolation not allowed'):
            sp = spectrum.interpolate_spectral_element(self.fname_ramp, -5)
        np.testing.assert_array_equal(sp(sp.waveset), 0)
        assert sp.warnings['DefaultThroughput']

    def test_exceptions(self):
        # Invalid filename format
        with pytest.raises(synexceptions.SynphotError):
            spectrum.interpolate_spectral_element('dummy.fits', -5)

        # Invalid interpolation column name
        with pytest.raises(synexceptions.SynphotError):
            spectrum.interpolate_spectral_element(
                irafconvert('cracscomp$acs_wfc_aper_002_syn.fits[mjd#]'),
                51252)

        # Cannot extrapolate and no default throughput
        with pytest.raises(synexceptions.ExtrapolationNotAllowed):
            spectrum.interpolate_spectral_element(self.fname_acs, -5)


@pytest.mark.remote_data
class TestObservationSpectralElement:
    """Test ``ObservationSpectralElement`` and ``band()``."""
    def setup_class(self):
        self.outdir = tempfile.mkdtemp()
        self.obs = spectrum.band(
            'acs,hrc,f555w', graphtable=GT_FILE, comptable=CP_FILE)
        self.sp = SourceSpectrum(ConstFlux1D, amplitude=1 * units.FLAM)

    def test_attributes(self):
        assert (str(self.obs.obsmode) == self.obs.meta['expr'] ==
                'acs,hrc,f555w')
        assert len(self.obs) == 6
        assert self.obs.area.value == conf.area
        np.testing.assert_array_equal(
            self.obs.waveset[::self.obs.waveset.size - 1].value, [500, 30010])
        np.testing.assert_array_equal(
            self.obs.binset[::self.obs.binset.size - 1].value, [1000, 11000])

    def test_config_primary_area(self):
        """Changing config after init should have no effect"""
        conf.area = 1
        assert conf.area == 1 and self.obs.area.value != 1
        conf.reset('area')

    def test_no_zero_bound(self):
        with pytest.warns(UserWarning, match=r'Unbounded throughput'):
            result = self.obs.bounded_by_zero(wavelengths=[5000, 6000])
        assert not result

    def test_other_graph_table(self):
        """Using the graph table with PRIMAREA."""
        gt_file = get_pkg_data_filename(
            os.path.join('data', 'tables_primarea_tmg.fits'))
        obs = spectrum.band(
            'acs,hrc,f555w', graphtable=gt_file, comptable=CP_FILE)

        # Primary area should be from graph table
        assert obs.area.value == 100

        x = obs.waveset.value
        y = obs(obs.waveset).value

        # Unit response
        a = obs.unit_response(obs.area)
        b = units.HC * 0.01 / abs(np.trapz(x * y, x=x))
        np.testing.assert_allclose(a.value, b.value)

    @pytest.mark.parametrize(
        ('obsmode', 'ans'),
        [('acs,hrc,f555w', 2.9788972414188295e-19),
         ('acs,sbc,f125lp', 1.7218083497870695e-17),
         ('acs,wfc1,f555w,f814w', 1.7485647715025005e-13),
         ('cos,boa,fuv,g130m,c1309', 3.8108219824401987e-15),
         ('stis,ccd,f25ndq1,a2d4,mjd#55555', 3.0597787848106823e-18),
         ('wfc3,ir,f140w', 1.4737148727216957e-20),
         ('wfc3,uvis1,f395n', 5.9433862614148255e-18),
         ('wfc3,uvis2,fq924n', 6.1632709911742462e-18),
         ('wfpc2,1,a2d7,f300w,cont#49892.0', 6.3011E-17),
         ('wfpc2,f555w', 4.8967453103320938e-19)])
    def test_uresp(self, obsmode, ans):
        """Unit response for different detector settings."""
        with warnings.catch_warnings():
            warnings.filterwarnings(
                'ignore', message=r'.*not valid flux unit.*',
                category=AstropyUserWarning)
            obs = spectrum.band(
                obsmode, graphtable=GT_FILE, comptable=CP_FILE)

        np.testing.assert_allclose(
            obs.unit_response(obs.area).value, ans, rtol=1e-4)

    def test_thermback(self):
        obs = spectrum.band(
            'wfc3,ir,f153m', graphtable=GT_FILE, comptable=CP_FILE)
        bg = obs.thermback(thermtable=TH_FILE)
        np.testing.assert_allclose(
            bg, 5.9774451061328011e-2 * (u.count / u.s / u.pix), rtol=5e-3)

        # ACS has no thermal background
        with pytest.raises(NotImplementedError):
            bg = self.obs.thermback(thermtable=TH_FILE)

    @pytest.mark.parametrize(
        ('cenwave', 'ans'),
        [(500 * u.nm, [499.9, 500.1] * u.nm),
         (5000, [4999, 5001] * u.AA)])
    def test_binned_waverange(self, cenwave, ans):
        np.testing.assert_allclose(
            self.obs.binned_waverange(cenwave, 2, mode='none'), ans)

    def test_binned_pixelrange(self):
        assert self.obs.binned_pixelrange(
            [499.95, 500.05] * u.nm, mode='round') == 1

    def test_std_filter(self):
        obs1 = spectrum.band('johnson,v')
        obs2 = SpectralElement.from_filter('johnson_v', encoding='binary')
        w = obs1.waveset
        np.testing.assert_allclose(obs1(w), obs2(w))

        # No pixel scale
        with pytest.raises(PixscaleNotFoundError):
            obs1.thermback()

        # No binset
        with pytest.raises(synexceptions.UndefinedBinset):
            obs1.binned_waverange(5000, 2)
        with pytest.raises(synexceptions.UndefinedBinset):
            obs1.binned_pixelrange([5000, 5002])

    def test_no_obsmode(self):
        """Spectrum without obsmode."""
        with pytest.raises(synexceptions.SynphotError):
            spectrum.ObservationSpectralElement(Const1D, amplitude=1)

    def test_write_fits(self):
        outfile = os.path.join(self.outdir, 'outspec1.fits')
        self.obs.to_fits(outfile, trim_zero=False, pad_zero_ends=False)

        # Read it back in
        with fits.open(outfile) as pf:
            assert (pf[1].header['grftable'] ==
                    os.path.basename(self.obs.obsmode.gtname))
            assert (pf[1].header['cmptable'] ==
                    os.path.basename(self.obs.obsmode.ctname))

        obs = SpectralElement.from_file(outfile)
        w = self.obs.waveset
        np.testing.assert_allclose(obs.waveset, w)
        np.testing.assert_allclose(obs(w), self.obs(w))

    def test_disabled_methods(self):
        with pytest.raises(NotImplementedError):
            self.obs.taper()
        with pytest.raises(NotImplementedError):
            self.obs.from_file('dummy.fits')
        with pytest.raises(NotImplementedError):
            self.obs.from_filter('johnson_v')

    def teardown_class(self):
        shutil.rmtree(self.outdir)


@pytest.mark.remote_data
class TestEbmvx:
    """Test extinction curve and related cache."""
    def setup_class(self):
        self.ec_mwavg = spectrum.ebmvx('mwavg', 0.3)

        # https://github.com/spacetelescope/synphot_refactor/issues/129
        self.w = np.squeeze(self.ec_mwavg.model.points)

        self.y = self.ec_mwavg(self.w)

    def test_mwavg(self):
        """No check on data quality, which is dependent on reference file."""
        assert spectrum._REDLAWS['mwavg'].meta['expr'] == 'mwavg'

    @pytest.mark.parametrize('m', ['gal3', None])
    def test_dummy(self, m):
        """Dummy values should default to 'mwavg'."""
        ec_test = spectrum.ebmvx(m, 0.3)
        np.testing.assert_array_equal(ec_test(self.w), self.y)

    def teardown_class(self):
        spectrum.reset_cache()
        assert spectrum._REDLAWS == {}


def test_vega_dummy():
    """Test that Vega spectrum is loaded properly."""
    # Dummy
    with pytest.warns(AstropyUserWarning,
                      match=r'Failed to load Vega spectrum'):
        spectrum.load_vega(vegafile='dummyfile.fits', encoding='binary')
    assert spectrum.Vega is None


@pytest.mark.remote_data
def test_vega_default():
    """Test that Vega spectrum is loaded properly."""
    # Default
    spectrum.load_vega(encoding='binary')
    assert 'Vega' in spectrum.Vega.meta['expr']



"""This module contains ACS/WFC commissioning tests.
Adapted from ``astrolib/pysynphot/from_commissioning/acs/test5.py``,
``astrolib/pysynphot/from_commissioning/acs/test6.py``, and
``astrolib/pysynphot/from_commissioning/acs/test7.py``.

.. note::

    ``astrolib/pysynphot/from_commissioning/acs/test4.py`` was disabled.
    Therefore, Test658 to Test671 were not ported over.

    WFC1 and WFC2 have the same throughput curves.
    Therefore, testing for WFC1 alone is sufficient.

"""

# STDLIB
import os
import shutil

# THIRD-PARTY
from astropy.utils.data import get_pkg_data_filename

# LOCAL
from ..utils import CommCase

# Local test data
datafiles = ['earthshine.fits', 'el1215a.fits', 'el1302a.fits', 'el1356a.fits',
             'el2471a.fits', 'Zodi.fits']


def setup_module(module):
    """Copy test data to working directory so ``parse_spec`` works
    properly for both ``stsynphot`` and ASTROLIB PYSYNPHOT."""
    for datafile in datafiles:
        src = get_pkg_data_filename(os.path.join('data', datafile))
        shutil.copyfile(src, datafile)


def teardown_module(module):
    """Clean up test data in working directory."""
    for datafile in datafiles:
        os.remove(datafile)


class Test672(CommCase):
    obsmode = 'acs,wfc1,f502n'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test673(CommCase):
    obsmode = 'acs,wfc1,f550m'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test674(Test673):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test675(Test673):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test676(Test673):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test677(Test673):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test678(CommCase):
    obsmode = 'acs,wfc1,f555w'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test679(Test678):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test680(Test678):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test681(Test678):
    spectrum = 'rn(bb(10000),band(johnson,v),20,vegamag)'


class Test682(Test678):
    spectrum = 'rn(icat(k93models,5770,0.0,4.5),band(johnson,v),20,vegamag)'


class Test683(Test678):
    spectrum = 'rn(pl(4000.0,-1.0,flam),band(johnson,v),20,vegamag)'


class Test684(Test678):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test685(Test678):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),22,vegamag)'


class Test686(Test678):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


# Duplicate of Test680 and syntax already tested in Test490
# class Test687(Test678):
#    spectrum = 'spec($PYSYN_CDBS/calspec/gd71_mod_005.fits)'


class Test688(Test678):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test689(Test678):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test690(CommCase):
    obsmode = 'acs,wfc1,f555w,pol_v'
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),22,vegamag)'


class Test691(Test690):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test692(CommCase):
    obsmode = 'acs,wfc1,f606w'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test693(Test692):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test694(Test692):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test695(Test692):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test696(Test692):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test697(CommCase):
    obsmode = 'acs,wfc1,f625w'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test698(Test697):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test699(Test697):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test700(Test697):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test701(Test697):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test702(Test697):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test703(CommCase):
    obsmode = 'acs,wfc1,f658n'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test704(Test703):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test705(Test703):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test706(Test703):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test707(Test703):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test708(CommCase):
    obsmode = 'acs,wfc1,f660n'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test709(Test708):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test710(Test708):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test711(Test708):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test712(Test708):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test713(CommCase):
    obsmode = 'acs,wfc1,f775w'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test714(Test713):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test715(Test713):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test716(Test713):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test717(Test713):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test718(CommCase):
    obsmode = 'acs,wfc1,f814w'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test719(Test718):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test720(Test718):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test721(Test718):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test722(Test718):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test723(CommCase):
    obsmode = 'acs,wfc1,f850lp'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test724(Test723):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test725(Test723):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test726(Test723):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test727(Test723):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test728(Test723):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test729(CommCase):
    obsmode = 'acs,wfc1,f892n'
    spectrum = '$PYSYN_CDBS/calspec/g191b2b_mod_004.fits'


class Test730(Test729):
    spectrum = '$PYSYN_CDBS/calspec/gd153_mod_004.fits'


class Test731(Test729):
    spectrum = '$PYSYN_CDBS/calspec/gd71_mod_005.fits'


class Test732(Test729):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1e-18,flam)'


class Test733(Test729):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')


class Test734(CommCase):
    obsmode = 'acs,wfc1,fr1016n#10000'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test735(Test734):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test736(CommCase):
    obsmode = 'acs,wfc1,fr388n#3880'
    spectrum = 'em(3880.0,10.0,1.0E-16,flam)'


class Test737(Test736):
    spectrum = 'rn(bb(10000),band(johnson,v),20,vegamag)'


class Test738(Test736):
    spectrum = 'rn(icat(k93models,15400,0.0,3.9),band(johnson,v),15,vegamag)'


class Test739(Test736):
    spectrum = 'rn(icat(k93models,3500,0.0,4.6),band(johnson,v),15,vegamag)'


class Test740(Test736):
    spectrum = 'rn(icat(k93models,44500,0.0,5.0),band(johnson,v),15,vegamag)'


class Test741(Test736):
    spectrum = 'rn(icat(k93models,4850,0.0,1.1),band(johnson,v),15,vegamag)'


class Test742(Test736):
    spectrum = 'rn(icat(k93models,5770,0.0,4.5),band(johnson,v),15,vegamag)'


class Test743(Test736):
    spectrum = 'rn(icat(k93models,5770,0.0,4.5),band(johnson,v),20,vegamag)'


class Test744(Test736):
    spectrum = 'rn(pl(4000.0,-1.0,flam),band(johnson,v),20,vegamag)'


class Test745(Test736):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test746(Test736):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),22,vegamag)'


# Duplicate of Test748 and syntax already tested in Test587
# class Test747(Test736):
#    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.E-15,flam)'


class Test748(Test736):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test749(Test736):
    spectrum = 'spec($PYSYN_CDBS/calspec/g191b2b_mod_004.fits)'


class Test750(Test736):
    spectrum = 'spec($PYSYN_CDBS/calspec/gd71_mod_005.fits)'


class Test751(Test736):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test752(CommCase):
    obsmode = 'acs,wfc1,fr388n#3881'
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test753(Test752):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test754(CommCase):
    obsmode = 'acs,wfc1,fr423n#4230'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test755(Test754):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test756(CommCase):
    obsmode = 'acs,wfc1,fr459m#4590'
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),22,vegamag)'


class Test757(Test756):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test758(CommCase):
    obsmode = 'acs,wfc1,fr459m#4620'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test759(Test758):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test760(CommCase):
    obsmode = 'acs,wfc1,fr462n#4620'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test761(Test760):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test762(CommCase):
    obsmode = 'acs,wfc1,fr505n#5000'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test763(Test762):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test764(CommCase):
    obsmode = 'acs,wfc1,fr551n#5500'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test765(Test764):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test766(CommCase):
    obsmode = 'acs,wfc1,fr601n#6000'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test767(Test766):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test768(CommCase):
    obsmode = 'acs,wfc1,fr647m#6470'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test769(Test768):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test770(CommCase):
    obsmode = 'acs,wfc1,fr656n#6500'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test771(Test770):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test772(CommCase):
    obsmode = 'acs,wfc1,fr716n#7100'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test773(Test772):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test774(CommCase):
    obsmode = 'acs,wfc1,fr782n#7900'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test775(Test774):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test776(CommCase):
    obsmode = 'acs,wfc1,fr853n#8500'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test777(Test776):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test778(CommCase):
    obsmode = 'acs,wfc1,fr914m#9000'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test779(Test778):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test780(CommCase):
    obsmode = 'acs,wfc1,fr931n#9300'
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.e-15,flam)'


class Test781(Test780):
    obsmode = 'acs,wfc1,fr931n#9300'
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test782(CommCase):
    obsmode = 'acs,wfc1,g800l'
    spectrum = 'em(6500.0,10.0,1.0E-16,flam)'


class Test783(Test782):
    spectrum = 'rn(bb(10000),band(johnson,v),20,vegamag)'


class Test784(Test782):
    spectrum = 'rn(icat(k93models,5770,0.0,4.5),band(johnson,v),20,vegamag)'


class Test785(Test782):
    spectrum = 'rn(pl(4000.0,-1.0,flam),band(johnson,v),20,vegamag)'


class Test786(Test782):
    spectrum = 'rn(unit(1.0,flam),band(johnson,v),15,vegamag)'


class Test787(Test782):
    spectrum = 'rn(unit(1.0,flam),box(5500.0,1.0),1.5e-16,flam)'


class Test788(Test782):
    spectrum = 'spec($PYSYN_CDBS/calspec/gd71_mod_005.fits)'


class Test789(Test782):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)')


class Test790(Test782):
    spectrum = ('spec(earthshine.fits)*0.5+rn(spec(Zodi.fits),'
                'band(johnson,v),22.7,vegamag)+(spec(el1215a.fits)+'
                'spec(el1302a.fits)+spec(el1356a.fits)+spec(el2471a.fits))')



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module handles graph and component (optical or thermal) tables."""

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import log

# LOCAL
from . import exceptions, stio
from .config import conf
from .stio import get_latest_file, irafconvert

__all__ = ['GraphTable', 'CompTable']


class GraphTable:
    """Class to handle graph table.

    Table is parsed with :func:`~stsynphot.stio.read_graphtable`.
    All string entries will be converted to lower case.
    Comment column is ignored.

    Parameters
    ----------
    graphfile : str
        Graph table name.

    ext : int, optional
        FITS extension index of the data table.

    Attributes
    ----------
    keywords : array of str
        Keyword names.

    compnames, thcompnames : array of str
        Components names (optical and thermal).

    innodes, outnodes : array of int
        Input and output nodes.

    primary_area : `astropy.units.quantity.Quantity` or `None`
        Value of PRIMAREA keyword in primary header.

    """
    def __init__(self, graphfile, ext=1):
        self.primary_area, data = stio.read_graphtable(
            get_latest_file(
                irafconvert(graphfile),
                err_msg=('No graph tables found; functionality will be '
                         'SEVERELY crippled.')),
            tab_ext=ext)

        # Convert all strings to lowercase
        self.keywords = np.array([s.lower() for s in data['KEYWORD']])
        self.compnames = np.array([s.lower() for s in data['COMPNAME']])
        self.thcompnames = np.array([s.lower() for s in data['THCOMPNAME']])

        # Already int
        self.innodes = data['INNODE']
        self.outnodes = data['OUTNODE']

    def get_next_node(self, modes, innode):
        """Return the output node that matches an element from
        given list of modes, starting at the given input node.

        If no match found for the given modes, output node
        corresponding to default mode is used.
        If multiple matches are found, only the result for the
        latest matched mode is stored.

        .. note::

            This is only used for debugging.

        Parameters
        ----------
        modes : list of str
            List of modes.

        innode : int
            Starting input node.

        Returns
        -------
        outnode : int
            Matching output node, or -1 if given input node not found.

        """
        nodes = np.where(self.innodes == innode)[0]

        # No match
        if len(nodes) == 0:
            return -1

        # Output node for default mode
        defaultindex = np.where(self.keywords[nodes] == 'default')[0]

        if len(defaultindex) != 0:
            outnode = self.outnodes[nodes[defaultindex]]

        for mode in modes:
            index = np.where(self.keywords[nodes] == mode)[0]
            if len(index) > 0:
                outnode = self.outnodes[nodes[index]]

        return outnode[0]

    def get_comp_from_gt(self, modes, innode):
        """Return component names for the given modes by traversing
        the graph table, starting at the given input node.

        .. note::

            Extra debug messages available by setting logger to
            debug mode.

        Parameters
        ----------
        modes : list of str
            List of modes.

        innode : int
            Starting input node.

        Returns
        -------
        components, thcomponents : list of str
            Optical and thermal components.

        Raises
        ------
        stsynphot.exceptions.AmbiguousObsmode
            Ambiguous mode.

        stsynphot.exceptions.IncompleteObsmode
            Incomplete mode.

        stsynphot.exceptions.UnusedKeyword
            Unused keyword in mode.

        """
        components = []
        thcomponents = []
        outnode = 0
        inmodes = set(modes)
        used_modes = set()
        count = 0

        while outnode >= 0:
            if outnode < 0:  # pragma: no cover
                log.debug(f'outnode={outnode} (stop condition).')

            previous_outnode = outnode
            nodes = np.where(self.innodes == innode)

            # If there are no entries with this innode, we're done
            if len(nodes[0]) == 0:
                log.debug(f'innode={innode} not found (stop condition).')
                break

            # Find the entry corresponding to the component named
            # 'default', because thats the one we'll use if we don't
            # match anything in the modes list
            if 'default' in self.keywords[nodes]:
                dfi = np.where(self.keywords[nodes] == 'default')[0][0]
                outnode = self.outnodes[nodes[0][dfi]]
                component = self.compnames[nodes[0][dfi]]
                thcomponent = self.thcompnames[nodes[0][dfi]]
                used_default = True
            else:
                # There's no default, so fail if nothing found in the
                # keyword matching step.
                outnode = -2
                component = thcomponent = None

            # Match something from the modes list
            for mode in modes:
                if mode in self.keywords[nodes]:
                    used_modes.add(mode)
                    index = np.where(self.keywords[nodes] == mode)
                    n_match = len(index[0])
                    if n_match > 1:
                        raise exceptions.AmbiguousObsmode(
                            f'{n_match} matches found for {mode}')
                    idx = index[0][0]
                    component = self.compnames[nodes[0][idx]]
                    thcomponent = self.thcompnames[nodes[0][idx]]
                    outnode = self.outnodes[nodes[0][idx]]
                    used_default = False

            log.debug(f'innode={innode} outnode={outnode} '
                      f'compname={component}')
            components.append(component)
            thcomponents.append(thcomponent)
            innode = outnode

            if outnode == previous_outnode:
                log.debug(f'innode={innode} outnode={outnode} '
                          f'used_default={used_default}')
                count += 1
                if count > 3:
                    log.debug(f'Same outnode={outnode} over 3 times (stop '
                              'condition)')
                    break

        if outnode < 0:
            log.debug(f'outnode={outnode} (stop condition)')
            raise exceptions.IncompleteObsmode(
                f'{modes}, choose from {self.keywords[nodes]}')

        if inmodes != used_modes:
            raise exceptions.UnusedKeyword(
                f'{str(inmodes.difference(used_modes))}')

        return components, thcomponents


class CompTable:
    """Class to handle component table (optical or thermal).

    Table is parsed with :func:`~stsynphot.stio.read_comptable`.
    Only component names and filenames are kept.
    Component throughput filenames are parsed with
    :func:`~stsynphot.stio.irafconvert`.

    Parameters
    ----------
    compfile : str
        Component table filename.

    ext : int, optional
        FITS extension index of the data table.

    Attributes
    ----------
    name : str
        Component table filename.

    compnames, filenames : array of str
        Component names and corresponding filenames.

    """
    def __init__(self, compfile, ext=1):
        data = stio.read_comptable(
            get_latest_file(
                irafconvert(compfile),
                err_msg=('No component tables found; functionality will be '
                         'SEVERELY crippled.')),
            tab_ext=ext)
        self.name = compfile
        self.compnames = np.array([s.lower() for s in data['COMPNAME']])
        self.filenames = np.array(
            list(map(stio.irafconvert, data['FILENAME'])))

    def get_filenames(self, compnames):
        """Get filenames of given component names.

        For multiple matches, only the first match is kept.

        Parameters
        ----------
        compnames : list of str
            List of component names to search. Case-sensitive.

        Returns
        -------
        files : list of str
            List of matched filenames.

        Raises
        ------
        stsynphot.exceptions.GraphtabError
            Unmatched component name.

        """
        files = []

        for compname in compnames:
            if compname not in (None, '', conf.clear_filter):
                index = np.where(self.compnames == compname)[0]
                if len(index) < 1:
                    raise exceptions.GraphtabError(
                        f'Cannot find {compname} in {self.name}.')
                files.append(self.filenames[index[0]].lstrip())
            else:
                files.append(conf.clear_filter)

        return files



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This sub-package is for commisioning tests.

.. note::

    Tests require ASTROLIB PYSYNPHOT to be installed.
    Specifically, install the version to be compared against ``stsynphot``.

To run only the commissioning tests from source checkout::

    cd ../..
    pytest --pyargs stsynphot/commissioning --remote-data --slow

Like above, but also use ``pytest-html`` plugin for a detail HTML report::

    pytest --pyargs stsynphot/commissioning --remote-data --slow --html=/full/path/report.html

To rerun only failed tests from the last run, add ``--lf`` in the list of
``args`` above. This option requires ``pytest-cache`` (enabled by default)
and is useful for debugging.

"""  # noqa

from . import utils  # noqa



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module handles :ref:`catalog spectra <stsynphot-spec-atlas>`."""

# STDLIB
import numbers

# THIRD-PARTY
import numpy as np

# ASTROPY
from astropy import units as u
from astropy.io import fits

# SYNPHOT
from synphot import exceptions as synexceptions
from synphot import units
from synphot.spectrum import SourceSpectrum
from synphot.utils import validate_totalflux

# LOCAL
from . import exceptions, stio

__all__ = ['reset_cache', 'get_catalog_index', 'grid_to_spec',
           'find_valid_g_phoenix', 'plot_phoenix']

_PARAM_NAMES = ['T_eff', 'metallicity', 'log_g']
_CACHE = {}  # Stores grid look-up parameters to reduce file I/O.


def reset_cache():
    """Empty the catalog grid cache."""
    global _CACHE
    _CACHE.clear()


def _par_from_parser(x):
    """Convert parser string to parameter value."""
    if not isinstance(x, (numbers.Real, u.Quantity)):
        x = float(x)
    return x


def _break_list(in_list, index, parameter):
    """Break input list into upper and lower lists."""
    array = np.array([parameters[index] for parameters in in_list],
                     dtype=np.float64)
    upper_array = array[array >= parameter]
    lower_array = array[array <= parameter]

    if upper_array.size == 0:
        raise exceptions.ParameterOutOfBounds(
            f"Parameter '{_PARAM_NAMES[index]}' exceeds data. "
            f"Max allowed={array.max()}, entered={parameter}.")
    if lower_array.size == 0:
        raise exceptions.ParameterOutOfBounds(
            f"Parameter '{_PARAM_NAMES[index]}' exceeds data. "
            f"Min allowed={array.min()}, entered={parameter}.")

    upper = upper_array.min()
    lower = lower_array.max()
    upper_list = []
    lower_list = []

    for i, parameters in enumerate(in_list):
        if array[i] >= parameter and array[i] <= upper:
            upper_list.append(parameters)
        if array[i] >= lower and array[i] <= parameter:
            lower_list.append(parameters)

    return upper_list, lower_list


def _get_spectrum(parlist, catdir):
    """Get list of spectra for given parameter list and base name."""
    name = parlist[3]

    filename = name.split('[')[0]
    column = name.split('[')[1][:-1]

    filename = stio.resolve_filename(catdir, *filename.split('/'))
    sp = SourceSpectrum.from_file(filename, flux_col=column)

    totflux = sp.integrate()
    try:
        validate_totalflux(totflux)
    except synexceptions.SynphotError:
        raise exceptions.ParameterOutOfBounds(
            f"Parameter '{parlist}' has no valid data.")

    result = [member for member in parlist]
    result.pop()
    result.append(sp)

    return result


def _interpolate_spectrum(sp1, sp2, par):
    """Interpolate spectra to the given parameter value."""
    spectrum1 = sp1.pop()
    spectrum2 = sp2.pop()
    par1 = sp1.pop()
    par2 = sp2.pop()

    if par1 == par2:
        sp = spectrum1
    else:
        a = (par1 - par) / (par1 - par2)
        b = 1.0 - a
        sp = a * spectrum2 + b * spectrum1

    result = [member for member in sp1]
    result.append(sp)

    return result


def get_catalog_index(gridname):
    """Extract catalog index (grid parameters).

    It is read once and then cached until the cache is cleared explicitly using
    :func:`reset_cache`.

    Parameters
    ----------
    gridname : str
        See :func:`grid_to_spec`.

    Returns
    -------
    cat_index : list
        List of ``[t_eff, metallicity, log_g, filename]``.

    catdir : str
        Directory containing the requested catalog.

    """
    if gridname == 'ck04models':
        catdir = 'crgridck04$'
    elif gridname == 'k93models':
        catdir = 'crgridk93$'
    elif gridname == 'phoenix':
        catdir = 'crgridphoenix$'
    else:
        raise synexceptions.SynphotError(
            f'{gridname} is not a supported catalog grid.')

    catdir = stio.irafconvert(catdir)
    filename = stio.resolve_filename(catdir, 'catalog.fits')

    # If not cached, read from grid catalog and cache it
    if filename not in _CACHE:
        data = stio.read_catalog(filename)  # EXT 1
        _CACHE[filename] = [list(map(float, index.split(','))) +
                            [data['FILENAME'][i]]
                            for i, index in enumerate(data['INDEX'])]

    return _CACHE[filename], catdir


def grid_to_spec(gridname, t_eff, metallicity, log_g):
    """Extract spectrum from given catalog grid parameters.
    Interpolate if necessary.

    Grid parameters are read with :func:`get_catalog_index`.

    Parameters
    ----------
    gridname : {'ck04models', 'k93models', 'phoenix'}
        Model to use:
            * ``ck04models`` - Castelli & Kurucz (2004)
            * ``k93models`` - Kurucz (1993)
            * ``phoenix`` - Allard et al. (2009)

    t_eff : str, float or `astropy.units.quantity.Quantity`
        Effective temperature of model.
        If not Quantity, assumed to be in Kelvin.
        If string (from parser), convert to Quantity.

    metallicity : str or float
        Metallicity of model.
        If string (from parser), convert to float.

    log_g : str or float
        Log surface gravity for model.
        If string (from parser), convert to float.

    Returns
    -------
    sp : `synphot.spectrum.SourceSpectrum`
        Empirical source spectrum.

    Raises
    ------
    stsynphot.exceptions.ParameterOutOfBounds
        Grid parameter out of bounds.

    synphot.exceptions.SynphotError
        Invalid inputs.

    """
    indices, catdir = get_catalog_index(gridname)

    metallicity = _par_from_parser(metallicity)
    if isinstance(metallicity, u.Quantity):
        raise synexceptions.SynphotError(
            'Quantity is not supported for metallicity.')

    log_g = _par_from_parser(log_g)
    if isinstance(log_g, u.Quantity):
        raise synexceptions.SynphotError(
            'Quantity is not supported for log surface gravity.')

    t_eff = units.validate_quantity(_par_from_parser(t_eff), u.K).value

    list0, list1 = _break_list(indices, 0, t_eff)

    list2, list3 = _break_list(list0, 1, metallicity)
    list4, list5 = _break_list(list1, 1, metallicity)

    list6, list7 = _break_list(list2, 2, log_g)
    list8, list9 = _break_list(list3, 2, log_g)
    list10, list11 = _break_list(list4, 2, log_g)
    list12, list13 = _break_list(list5, 2, log_g)

    sp1 = _get_spectrum(list6[0], catdir)
    sp2 = _get_spectrum(list7[0], catdir)
    sp3 = _get_spectrum(list8[0], catdir)
    sp4 = _get_spectrum(list9[0], catdir)
    sp5 = _get_spectrum(list10[0], catdir)
    sp6 = _get_spectrum(list11[0], catdir)
    sp7 = _get_spectrum(list12[0], catdir)
    sp8 = _get_spectrum(list13[0], catdir)

    spa1 = _interpolate_spectrum(sp1, sp2, log_g)
    spa2 = _interpolate_spectrum(sp3, sp4, log_g)
    spa3 = _interpolate_spectrum(sp5, sp6, log_g)
    spa4 = _interpolate_spectrum(sp7, sp8, log_g)

    spa5 = _interpolate_spectrum(spa1, spa2, metallicity)
    spa6 = _interpolate_spectrum(spa3, spa4, metallicity)

    spa7 = _interpolate_spectrum(spa5, spa6, t_eff)

    sp = spa7[0]
    sp.meta['expr'] = (f'{gridname}(T_eff={t_eff:g},'
                       f'metallicity={metallicity:g},log_g={log_g:g})')

    return sp


# NOTE: The following functions need refactoring if you want to generalize
# them to work on all the supported catalogs. This is because the catalogs
# do not share the same header structure. Generalization requires reading all
# the spectra tables, not just header, which will have a significant
# performance impact.
# Also see https://github.com/spacetelescope/stsynphot_refactor/issues/45

def find_valid_g_phoenix():  # pragma: no cover
    """Find valid ``log_g`` values in the Phoenix catalog.

    .. note::

        Takes time to run because it has to find the gaps by
        parsing the headers of individual data files in the catalog.

    Returns
    -------
    valid_g : dict
        Dictionary mapping ``(t_eff, metallicity)`` to a list of valid
        ``log_g`` values.

    uniq_metallicity : list
        Unique metallicity values in the catalog.

    """
    indices, catdir = get_catalog_index('phoenix')
    specfiles = {}
    valid_g = {}
    possible_g = set()
    uniq_metallicity = set()

    for a in indices:
        key = tuple(a[:2])
        uniq_metallicity.add(a[1])
        possible_g.add(a[2])

        if key not in specfiles:
            specfiles[key] = a[3].split('[')[0]

    for key, spfile in specfiles.items():
        hdr = fits.getheader(stio.resolve_filename(
            catdir, *spfile.split('/')))['LOGG*']
        valid_g[key] = []
        for val in hdr.values():
            gval = float(val)
            if gval in possible_g:
                valid_g[key].append(gval)

    return valid_g, sorted(uniq_metallicity)


def plot_phoenix(filename=''):  # pragma: no cover
    """Visualize the Phoenix catalog index (grid parameters).

    .. note:: Uses ``matplotlib``.

    .. note::

        Takes time to run because it has to find the gaps by
        parsing the headers of individual data files in the catalog.

    Parameters
    ----------
    filename : str
        If provided, plot is saved to given filename.

    """
    import matplotlib.pyplot as plt

    valid_g, uniq_metallicity = find_valid_g_phoenix()
    n_plots = len(uniq_metallicity)
    ax_map = {}

    fig, axes = plt.subplots(nrows=3, ncols=4)
    axes = axes.flatten()
    n_axes = axes.size

    if n_plots > n_axes:
        raise ValueError(f'Need {n_plots} but only {n_axes} subplots created')

    for i, metallicity in enumerate(uniq_metallicity):
        ax_map[metallicity] = axes[i]

    for i in range(n_plots, n_axes):
        axes[i].remove()

    for key, log_g in valid_g.items():
        metallicity = key[1]
        t_eff = [key[0]] * len(log_g)
        ax = ax_map[metallicity]
        ax.plot(t_eff, log_g, 'b,')
        ax.set_xlabel(r'$T_{\mathrm{eff}}$')
        ax.set_ylabel(r'$\log g$')
        ax.set_title(f'[M/H] = {metallicity:.2f}')
        ax.axvline(0, ls='--', lw=0.5, c='k')

    plt.tight_layout()
    plt.draw()

    if filename:
        fig.savefig(filename)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *  # noqa
# ----------------------------------------------------------------------------

# STSYNPHOT UI
from .config import getref, showref, conf  # noqa
from .spectrum import band, ebmvx, Vega  # noqa
from .catalog import grid_to_spec  # noqa
from .spparser import parse_spec  # noqa
